--- os_freebsd.c.orig	Sun Sep  5 03:16:07 2004
+++ os_freebsd.c	Fri Aug 19 16:52:36 2005
@@ -18,6 +18,7 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <fcntl.h>
 #include <err.h>
 #include <camlib.h>
 #include <cam/scsi/scsi_message.h>
@@ -27,6 +28,7 @@
 #include <fcntl.h>
 #include <glob.h>
 #include <fcntl.h>
+#include <stddef.h>
 
 
 #include "config.h"
@@ -60,19 +62,23 @@
          "                                              (Enables SMART on first disk)\n\n"
          "  smartctl -t long /dev/ad0              (Executes extended disk self-test)\n\n"
          "  smartctl --attributes --log=selftest --quietmode=errorsonly /dev/ad0\n"
-         "                                      (Prints Self-Test & Attribute errors)\n"
-//         "  smartctl -a --device=3ware,2 /dev/sda\n"
-//         "          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
+         "                                      (Prints Self-Test & Attribute errors)\n\n"
+         "  smartctl -a --device=3ware,2 /dev/twa0\n"
+         "  smartctl -a --device=3ware,2 /dev/twe0\n"
+         "                              (Prints all SMART information for ATA disk on\n"
+         "                                 third port of first 3ware RAID controller)\n"
          );
 #else
   printf(
-         "  smartctl -a /dev/ad0                       (Prints all SMART information)\n"
-         "  smartctl -s on -o on -S on /dev/ad0         (Enables SMART on first disk)\n"
-         "  smartctl -t long /dev/ad0              (Executes extended disk self-test)\n"
+         "  smartctl -a /dev/ad0                       (Prints all SMART information)\n\n"
+         "  smartctl -s on -o on -S on /dev/ad0         (Enables SMART on first disk)\n\n"
+         "  smartctl -t long /dev/ad0              (Executes extended disk self-test)\n\n"
          "  smartctl -A -l selftest -q errorsonly /dev/ad0\n"
-         "                                      (Prints Self-Test & Attribute errors)\n"
-//         "  smartctl -a -d 3ware,2 /dev/sda\n"
-//         "          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
+         "                                      (Prints Self-Test & Attribute errors)\n\n"
+         "  smartctl -a -d 3ware,2 /dev/twa0\n"
+         "  smartctl -a -d 3ware,2 /dev/twe0\n"
+         "                              (Prints all SMART information for ATA disk on\n"
+         "                                 third port of first 3ware RAID controller)\n"
          );
 #endif
   return;
@@ -109,7 +115,11 @@
   }
 
   if (parse_ok == CONTROLLER_ATA) {
+#ifdef IOCATAREQUEST
+    if ((fdchan->device = open(dev,O_RDONLY))<0) {
+#else
     if ((fdchan->atacommand = open("/dev/ata",O_RDWR))<0) {
+#endif
       int myerror = errno;      //preserve across free call
       free (fdchan);
       errno = myerror;
@@ -118,9 +128,24 @@
   }
 
   if (parse_ok == CONTROLLER_3WARE_678K_CHAR) {
-    char buf[512];
-    sprintf(buf,"/dev/twe%d",fdchan->device);
-    if ((fdchan->atacommand = open(buf,O_RDWR))<0) {
+#ifdef IOCATAREQUEST
+    if ((fdchan->device = open(dev,O_RDWR))<0) {
+#else
+    if ((fdchan->atacommand = open(dev,O_RDWR))<0) {
+#endif
+      int myerror = errno; // preserver across free call
+      free(fdchan);
+      errno=myerror;
+      return -1;
+    }
+  }
+
+  if (parse_ok == CONTROLLER_3WARE_9000_CHAR) {
+#ifdef IOCATAREQUEST
+    if ((fdchan->device = open(dev,O_RDWR))<0) {
+#else
+    if ((fdchan->atacommand = open(dev,O_RDWR))<0) {
+#endif
       int myerror = errno; // preserver across free call
       free(fdchan);
       errno=myerror;
@@ -167,8 +192,13 @@
     free(fdchan->devname);
   
   // close device, if open
+#ifdef IOCATAREQUEST
+  if (fdchan->device)
+    failed=close(fdchan->device);
+#else
   if (fdchan->atacommand)
     failed=close(fdchan->atacommand);
+#endif
 
   if (fdchan->scsicontrol)
     failed=close(fdchan->scsicontrol);
@@ -221,7 +251,7 @@
 }
 
 int ata_command_interface(int fd, smart_command_set command, int select, char *data) {
-#ifndef ATAREQUEST
+#if !defined(ATAREQUEST) && !defined(IOCATAREQUEST)
   // sorry, but without ATAng, we can't do anything here
   printwarning(BAD_KERNEL,NULL);
   errno = ENOSYS;
@@ -229,7 +259,11 @@
 #else
   struct freebsd_dev_channel* con;
   int retval, copydata=0;
+#ifdef IOCATAREQUEST
+  struct ata_ioc_request request;
+#else
   struct ata_cmd iocmd;
+#endif
   unsigned char buff[512];
 
   // check that "file descriptor" is valid
@@ -238,89 +272,97 @@
 
   bzero(buff,512);
 
+#ifdef IOCATAREQUEST
+  bzero(&request,sizeof(struct ata_ioc_request));
+#else
   bzero(&iocmd,sizeof(struct ata_cmd));
+#endif
   bzero(buff,512);
+
+#ifndef IOCATAREQUEST
   iocmd.cmd=ATAREQUEST;
   iocmd.channel=con->channel;
   iocmd.device=con->device;
+#define request iocmd.u.request
+#endif
 
-  iocmd.u.request.u.ata.command=ATA_SMART_CMD;
-  iocmd.u.request.timeout=600;
+  request.u.ata.command=ATA_SMART_CMD;
+  request.timeout=600;
   switch (command){
   case READ_VALUES:
-    iocmd.u.request.u.ata.feature=ATA_SMART_READ_VALUES;
-    iocmd.u.request.u.ata.lba=0xc24f<<8;
-    iocmd.u.request.flags=ATA_CMD_READ;
-    iocmd.u.request.data=buff;
-    iocmd.u.request.count=512;
+    request.u.ata.feature=ATA_SMART_READ_VALUES;
+    request.u.ata.lba=0xc24f<<8;
+    request.flags=ATA_CMD_READ;
+    request.data=buff;
+    request.count=512;
     copydata=1;
     break;
   case READ_THRESHOLDS:
-    iocmd.u.request.u.ata.feature=ATA_SMART_READ_THRESHOLDS;
-    iocmd.u.request.u.ata.count=1;
-    iocmd.u.request.u.ata.lba=1|(0xc24f<<8);
-    iocmd.u.request.flags=ATA_CMD_READ;
-    iocmd.u.request.data=buff;
-    iocmd.u.request.count=512;
+    request.u.ata.feature=ATA_SMART_READ_THRESHOLDS;
+    request.u.ata.count=1;
+    request.u.ata.lba=1|(0xc24f<<8);
+    request.flags=ATA_CMD_READ;
+    request.data=buff;
+    request.count=512;
     copydata=1;
     break;
   case READ_LOG:
-    iocmd.u.request.u.ata.feature=ATA_SMART_READ_LOG_SECTOR;
-    iocmd.u.request.u.ata.lba=select|(0xc24f<<8);
-    iocmd.u.request.u.ata.count=1;
-    iocmd.u.request.flags=ATA_CMD_READ;
-    iocmd.u.request.data=buff;
-    iocmd.u.request.count=512;
+    request.u.ata.feature=ATA_SMART_READ_LOG_SECTOR;
+    request.u.ata.lba=select|(0xc24f<<8);
+    request.u.ata.count=1;
+    request.flags=ATA_CMD_READ;
+    request.data=buff;
+    request.count=512;
     copydata=1;
     break;
   case IDENTIFY:
-    iocmd.u.request.u.ata.command=ATA_IDENTIFY_DEVICE;
-    iocmd.u.request.flags=ATA_CMD_READ;
-    iocmd.u.request.data=buff;
-    iocmd.u.request.count=512;
+    request.u.ata.command=ATA_IDENTIFY_DEVICE;
+    request.flags=ATA_CMD_READ;
+    request.data=buff;
+    request.count=512;
     copydata=1;
     break;
   case PIDENTIFY:
-    iocmd.u.request.u.ata.command=ATA_IDENTIFY_PACKET_DEVICE;
-    iocmd.u.request.flags=ATA_CMD_READ;
-    iocmd.u.request.data=buff;
-    iocmd.u.request.count=512;
+    request.u.ata.command=ATA_IDENTIFY_PACKET_DEVICE;
+    request.flags=ATA_CMD_READ;
+    request.data=buff;
+    request.count=512;
     copydata=1;
     break;
   case ENABLE:
-    iocmd.u.request.u.ata.feature=ATA_SMART_ENABLE;
-    iocmd.u.request.u.ata.lba=0xc24f<<8;
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_ENABLE;
+    request.u.ata.lba=0xc24f<<8;
+    request.flags=ATA_CMD_CONTROL;
     break;
   case DISABLE:
-    iocmd.u.request.u.ata.feature=ATA_SMART_DISABLE;
-    iocmd.u.request.u.ata.lba=0xc24f<<8;
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_DISABLE;
+    request.u.ata.lba=0xc24f<<8;
+    request.flags=ATA_CMD_CONTROL;
     break;
   case AUTO_OFFLINE:
     // NOTE: According to ATAPI 4 and UP, this command is obsolete
-    iocmd.u.request.u.ata.feature=ATA_SMART_AUTO_OFFLINE;
-    iocmd.u.request.u.ata.lba=select|(0xc24f<<8);
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_AUTO_OFFLINE;
+    request.u.ata.lba=select|(0xc24f<<8);
+    request.flags=ATA_CMD_CONTROL;
     break;
   case AUTOSAVE:
-    iocmd.u.request.u.ata.feature=ATA_SMART_AUTOSAVE;
-    iocmd.u.request.u.ata.count=0xf1;  // to enable autosave
-    iocmd.u.request.u.ata.lba=0xc24f<<8;
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_AUTOSAVE;
+    request.u.ata.count=0xf1;  // to enable autosave
+    request.u.ata.lba=0xc24f<<8;
+    request.flags=ATA_CMD_CONTROL;
     break;
   case IMMEDIATE_OFFLINE:
-    iocmd.u.request.u.ata.feature=ATA_SMART_IMMEDIATE_OFFLINE;
-    iocmd.u.request.u.ata.lba = select|(0xc24f<<8); // put test in sector
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_IMMEDIATE_OFFLINE;
+    request.u.ata.lba = select|(0xc24f<<8); // put test in sector
+    request.flags=ATA_CMD_CONTROL;
     break;
   case STATUS_CHECK: // same command, no HDIO in FreeBSD
   case STATUS:
     // this command only says if SMART is working.  It could be
     // replaced with STATUS_CHECK below.
-    iocmd.u.request.u.ata.feature=ATA_SMART_STATUS;
-    iocmd.u.request.u.ata.lba=0xc24f<<8;
-    iocmd.u.request.flags=ATA_CMD_CONTROL;
+    request.u.ata.feature=ATA_SMART_STATUS;
+    request.u.ata.lba=0xc24f<<8;
+    request.flags=ATA_CMD_CONTROL;
     break;
   default:
     pout("Unrecognized command %d in ata_command_interface()\n"
@@ -334,15 +376,19 @@
     unsigned const char failed_lo=0xf4, failed_hi=0x2c;
     unsigned char low,high;
     
+#ifdef IOCATAREQUEST
+    if ((retval=ioctl(con->device, IOCATAREQUEST, &request)))
+#else
     if ((retval=ioctl(con->atacommand, IOCATA, &iocmd)))
+#endif
       return -1;
 
 #if __FreeBSD_version < 502000
     printwarning(NO_RETURN,NULL);
 #endif
 
-    high = (iocmd.u.request.u.ata.lba >> 16) & 0xff;
-    low = (iocmd.u.request.u.ata.lba >> 8) & 0xff;
+    high = (request.u.ata.lba >> 16) & 0xff;
+    low = (request.u.ata.lba >> 8) & 0xff;
     
     // Cyl low and Cyl high unchanged means "Good SMART status"
     if (low==normal_lo && high==normal_hi)
@@ -355,18 +401,22 @@
     // We haven't gotten output that makes sense; print out some debugging info
     char buf[512];
     sprintf(buf,"CMD=0x%02x\nFR =0x%02x\nNS =0x%02x\nSC =0x%02x\nCL =0x%02x\nCH =0x%02x\nRETURN =0x%04x\n",
-            (int)iocmd.u.request.u.ata.command,
-            (int)iocmd.u.request.u.ata.feature,
-            (int)iocmd.u.request.u.ata.count,
-            (int)((iocmd.u.request.u.ata.lba) & 0xff),
-            (int)((iocmd.u.request.u.ata.lba>>8) & 0xff),
-            (int)((iocmd.u.request.u.ata.lba>>16) & 0xff),
-            (int)iocmd.u.request.error);
+            (int)request.u.ata.command,
+            (int)request.u.ata.feature,
+            (int)request.u.ata.count,
+            (int)((request.u.ata.lba) & 0xff),
+            (int)((request.u.ata.lba>>8) & 0xff),
+            (int)((request.u.ata.lba>>16) & 0xff),
+            (int)request.error);
     printwarning(BAD_SMART,buf);
     return 0;   
   }
 
+#ifdef IOCATAREQUEST
+  if ((retval=ioctl(con->device, IOCATAREQUEST, &request))) {
+#else
   if ((retval=ioctl(con->atacommand, IOCATA, &iocmd))) {
+#endif
     perror("Failed command: ");
     return -1;
   }
@@ -488,11 +538,12 @@
   struct freebsd_dev_channel* con;
 
   // return value and buffer for ioctl()
-  int  ioctlreturn, readdata=0;
-  struct twe_usercommand* cmd = NULL;
+  int  ioctlreturn;
+  TWE_Command* cmd = NULL;
 
   // Used by both the SCSI and char interfaces
-  char ioctl_buffer[sizeof(struct twe_usercommand)];
+  void *ioctl_buffer;
+  size_t ioctl_buffer_length;
 
   if (disknum < 0) {
     printwarning(NO_DISK_3WARE,NULL);
@@ -503,27 +554,44 @@
   if (isnotopen(&fd,&con))
       return -1;
 
-  memset(ioctl_buffer, 0, sizeof(struct twe_usercommand));
+  if (escalade_type==CONTROLLER_3WARE_9000_CHAR)
+    ioctl_buffer_length = sizeof(TW_OSLI_IOCTL_NO_DATA_BUF);
+  else if (escalade_type==CONTROLLER_3WARE_678K_CHAR)
+    ioctl_buffer_length = sizeof(struct twe_usercommand);
+  else {
+    errno=ENOTSUP;
+    return -1;
+  }
+  ioctl_buffer = malloc(ioctl_buffer_length);
+  if (ioctl_buffer == NULL)
+    return -ENOMEM;
+
+  bzero(ioctl_buffer, ioctl_buffer_length);
 
-  cmd = (struct twe_usercommand*)ioctl_buffer;
-  cmd->tu_command.ata.opcode = TWE_OP_ATA_PASSTHROUGH;
+  if (escalade_type==CONTROLLER_3WARE_9000_CHAR)
+    cmd = (TWE_Command *)(ioctl_buffer +
+		offsetof(TW_OSLI_IOCTL_NO_DATA_BUF, cmd_pkt.command.cmd_pkt_7k));
+  else if (escalade_type==CONTROLLER_3WARE_678K_CHAR)
+    cmd = (TWE_Command *)(ioctl_buffer +
+		offsetof(struct twe_usercommand, tu_command));
 
   // Same for (almost) all commands - but some reset below
-  cmd->tu_command.ata.request_id    = 0xFF;
-  cmd->tu_command.ata.unit   = disknum;
-  cmd->tu_command.ata.host_id = 0;
-  cmd->tu_command.ata.status        = 0;           
-  cmd->tu_command.ata.flags         = 0x1;
-  cmd->tu_command.ata.drive_head    = 0x0;
-  cmd->tu_command.ata.sector_num    = 0;
+  cmd->ata.opcode         = TWE_OP_ATA_PASSTHROUGH;
+  cmd->ata.request_id     = 0xFF;
+  cmd->ata.unit           = disknum;
+  cmd->ata.host_id        = 0;
+  cmd->ata.status         = 0;           
+  cmd->ata.flags          = 0x1;
+  cmd->ata.drive_head     = 0x0;
+  cmd->ata.sector_num     = 0;
 
   // All SMART commands use this CL/CH signature.  These are magic
   // values from the ATA specifications.
-  cmd->tu_command.ata.cylinder_lo   = 0x4F;
-  cmd->tu_command.ata.cylinder_hi   = 0xC2;
+  cmd->ata.cylinder_lo    = 0x4F;
+  cmd->ata.cylinder_hi    = 0xC2;
   
   // SMART ATA COMMAND REGISTER value
-  cmd->tu_command.ata.command       = ATA_SMART_CMD;
+  cmd->ata.command       = ATA_SMART_CMD;
   
   // Is this a command that reads or returns 512 bytes?
   // passthru->param values are:
@@ -537,120 +605,135 @@
       command == READ_LOG        ||
       command == IDENTIFY        ||
       command == WRITE_LOG ) {
-    readdata=1;
-    cmd->tu_size = 512;
-    cmd->tu_data = data;
-    cmd->tu_command.ata.sgl_offset = 0x5;
-    cmd->tu_command.ata.size         = 0x5;
-    cmd->tu_command.ata.param        = 0xD;
-    cmd->tu_command.ata.sector_count = 0x1;
+    if (escalade_type==CONTROLLER_3WARE_9000_CHAR) {
+      TW_OSLI_IOCTL_NO_DATA_BUF *buf = (TW_OSLI_IOCTL_NO_DATA_BUF *)ioctl_buffer;
+      buf->driver_pkt.buffer_length = 512; /* payload size */
+      buf->pdata = data;
+      ((TW_OSLI_IOCTL_NO_DATA_BUF *)ioctl_buffer)->pdata = data;
+    } else if (escalade_type==CONTROLLER_3WARE_678K_CHAR) {
+      struct twe_usercommand *buf = (struct twe_usercommand *)ioctl_buffer;
+      buf->tu_size = 512; /* payload size */
+      buf->tu_data = data;
+    }
+    cmd->ata.sgl_offset   = 0x5;
+    cmd->ata.size         = 0x5;
+    cmd->ata.param        = 0xD;
+    cmd->ata.sector_count = 0x1;
     // For 64-bit to work correctly, up the size of the command packet
     // in dwords by 1 to account for the 64-bit single sgl 'address'
     // field. Note that this doesn't agree with the typedefs but it's
     // right (agree with kernel driver behavior/typedefs).
     //if (sizeof(long)==8)
-    //  cmd->tu_command.ata.size++;
+    //  cmd->ata.size++;
   }
   else {
     // Non data command -- but doesn't use large sector 
     // count register values.  
-    cmd->tu_command.ata.sgl_offset = 0x0;
-    cmd->tu_command.ata.size         = 0x5;
-    cmd->tu_command.ata.param        = 0x8;
-    cmd->tu_command.ata.sector_count = 0x0;
+    cmd->ata.sgl_offset   = 0x0;
+    cmd->ata.size         = 0x5;
+    cmd->ata.param        = 0x8;
+    cmd->ata.sector_count = 0x0;
   }
   
   // Now set ATA registers depending upon command
   switch (command){
   case CHECK_POWER_MODE:
-    cmd->tu_command.ata.command     = ATA_CHECK_POWER_MODE;
-    cmd->tu_command.ata.features    = 0;
-    cmd->tu_command.ata.cylinder_lo = 0;
-    cmd->tu_command.ata.cylinder_hi = 0;
+    cmd->ata.command      = ATA_CHECK_POWER_MODE;
+    cmd->ata.features     = 0;
+    cmd->ata.cylinder_lo  = 0;
+    cmd->ata.cylinder_hi  = 0;
     break;
   case READ_VALUES:
-    cmd->tu_command.ata.features = ATA_SMART_READ_VALUES;
+    cmd->ata.features     = ATA_SMART_READ_VALUES;
     break;
   case READ_THRESHOLDS:
-    cmd->tu_command.ata.features = ATA_SMART_READ_THRESHOLDS;
+    cmd->ata.features     = ATA_SMART_READ_THRESHOLDS;
     break;
   case READ_LOG:
-    cmd->tu_command.ata.features = ATA_SMART_READ_LOG_SECTOR;
+    cmd->ata.features     = ATA_SMART_READ_LOG_SECTOR;
     // log number to return
-    cmd->tu_command.ata.sector_num  = select;
+    cmd->ata.sector_num   = select;
     break;
   case WRITE_LOG:
-    cmd->tu_data = data;
-    readdata=0;
-    cmd->tu_command.ata.features     = ATA_SMART_WRITE_LOG_SECTOR;
-    cmd->tu_command.ata.sector_count = 1;
-    cmd->tu_command.ata.sector_num   = select;
-    cmd->tu_command.ata.param        = 0xF;  // PIO data write
+    cmd->ata.features     = ATA_SMART_WRITE_LOG_SECTOR;
+    cmd->ata.sector_count = 1;
+    cmd->ata.sector_num   = select;
+    cmd->ata.param        = 0xF;  // PIO data write
     break;
   case IDENTIFY:
     // ATA IDENTIFY DEVICE
-    cmd->tu_command.ata.command     = ATA_IDENTIFY_DEVICE;
-    cmd->tu_command.ata.features    = 0;
-    cmd->tu_command.ata.cylinder_lo = 0;
-    cmd->tu_command.ata.cylinder_hi = 0;
+    cmd->ata.command      = ATA_IDENTIFY_DEVICE;
+    cmd->ata.features     = 0;
+    cmd->ata.cylinder_lo  = 0;
+    cmd->ata.cylinder_hi  = 0;
     break;
   case PIDENTIFY:
     // 3WARE controller can NOT have packet device internally
     pout("WARNING - NO DEVICE FOUND ON 3WARE CONTROLLER (disk %d)\n", disknum);
     errno=ENODEV;
-    return -1;
+    goto failed;
   case ENABLE:
-    cmd->tu_command.ata.features = ATA_SMART_ENABLE;
+    cmd->ata.features     = ATA_SMART_ENABLE;
     break;
   case DISABLE:
-    cmd->tu_command.ata.features = ATA_SMART_DISABLE;
+    cmd->ata.features     = ATA_SMART_DISABLE;
     break;
   case AUTO_OFFLINE:
-    cmd->tu_command.ata.features     = ATA_SMART_AUTO_OFFLINE;
+    cmd->ata.features     = ATA_SMART_AUTO_OFFLINE;
     // Enable or disable?
-    cmd->tu_command.ata.sector_count = select;
+    cmd->ata.sector_count = select;
     break;
   case AUTOSAVE:
-    cmd->tu_command.ata.features     = ATA_SMART_AUTOSAVE;
+    cmd->ata.features     = ATA_SMART_AUTOSAVE;
     // Enable or disable?
-    cmd->tu_command.ata.sector_count = select;
+    cmd->ata.sector_count = select;
     break;
   case IMMEDIATE_OFFLINE:
-    cmd->tu_command.ata.features    = ATA_SMART_IMMEDIATE_OFFLINE;
+    cmd->ata.features     = ATA_SMART_IMMEDIATE_OFFLINE;
     // What test type to run?
-    cmd->tu_command.ata.sector_num  = select;
+    cmd->ata.sector_num   = select;
     break;
   case STATUS_CHECK:
-    cmd->tu_command.ata.features = ATA_SMART_STATUS;
+    cmd->ata.features     = ATA_SMART_STATUS;
     break;
   case STATUS:
     // This is JUST to see if SMART is enabled, by giving SMART status
     // command. But it doesn't say if status was good, or failing.
     // See below for the difference.
-    cmd->tu_command.ata.features = ATA_SMART_STATUS;
+    cmd->ata.features     = ATA_SMART_STATUS;
     break;
   default:
     pout("Unrecognized command %d in freebsd_3ware_command_interface(disk %d)\n"
          "Please contact " PACKAGE_BUGREPORT "\n", command, disknum);
     errno=ENOSYS;
-    return -1;
+    goto failed;
   }
 
   // Now send the command down through an ioctl()
-  ioctlreturn=ioctl(con->atacommand,TWEIO_COMMAND,cmd);
+#ifdef IOCATAREQUEST
+  if (escalade_type==CONTROLLER_3WARE_9000_CHAR)
+    ioctlreturn=ioctl(con->device,TW_OSL_IOCTL_FIRMWARE_PASS_THROUGH,ioctl_buffer);
+  else
+    ioctlreturn=ioctl(con->device,TWEIO_COMMAND,ioctl_buffer);
+#else
+  if (escalade_type==CONTROLLER_3WARE_9000_CHAR)
+    ioctlreturn=ioctl(con->atacommand,TW_OSL_IOCTL_FIRMWARE_PASS_THROUGH,ioctl_buffer);
+  else
+    ioctlreturn=ioctl(con->atacommand,TWEIO_COMMAND,ioctl_buffer);
+#endif
   
   // Deal with the different error cases
   if (ioctlreturn) {
     if (!errno)
       errno=EIO;
-    return -1;
+    goto failed;
   }
   
   // See if the ATA command failed.  Now that we have returned from
   // the ioctl() call, if passthru is valid, then:
-  // - cmd->tu_command.ata.status contains the 3ware controller STATUS
-  // - cmd->tu_command.ata.command contains the ATA STATUS register
-  // - cmd->tu_command.ata.features contains the ATA ERROR register
+  // - cmd->ata.status contains the 3ware controller STATUS
+  // - cmd->ata.command contains the ATA STATUS register
+  // - cmd->ata.features contains the ATA ERROR register
   //
   // Check bits 0 (error bit) and 5 (device fault) of the ATA STATUS
   // If bit 0 (error bit) is set, then ATA ERROR register is valid.
@@ -658,10 +741,11 @@
   // doesn't make much sense: we don't care in detail why the error
   // happened.
   
-  if (cmd->tu_command.ata.status || (cmd->tu_command.ata.command & 0x21)) {
-    pout("Command failed, ata.status=(0x%2.2x), ata.command=(0x%2.2x), ata.flags=(0x%2.2x)\n",cmd->tu_command.ata.status,cmd->tu_command.ata.command,cmd->tu_command.ata.flags);
+  if (cmd->ata.status || (cmd->ata.command & 0x21)) {
+    pout("Command failed, ata.status=(0x%2.2x), ata.command=(0x%2.2x), ata.flags=(0x%2.2x)\n",
+	 cmd->ata.status,cmd->ata.command,cmd->ata.flags);
     errno=EIO;
-    return -1;
+    goto failed;
   }
   
   // For STATUS_CHECK, we need to check register values
@@ -671,44 +755,60 @@
     // need to examine the values of the Cylinder Low and Cylinder
     // High Registers.
     
-    unsigned short cyl_lo=cmd->tu_command.ata.cylinder_lo;
-    unsigned short cyl_hi=cmd->tu_command.ata.cylinder_hi;
+    unsigned short cyl_lo=cmd->ata.cylinder_lo;
+    unsigned short cyl_hi=cmd->ata.cylinder_hi;
     
     // If values in Cyl-LO and Cyl-HI are unchanged, SMART status is good.
-    if (cyl_lo==0x4F && cyl_hi==0xC2)
+    if (cyl_lo==0x4F && cyl_hi==0xC2) {
+      free(ioctl_buffer);
       return 0;
+    }
     
     // If values in Cyl-LO and Cyl-HI are as follows, SMART status is FAIL
-    if (cyl_lo==0xF4 && cyl_hi==0x2C)
+    if (cyl_lo==0xF4 && cyl_hi==0x2C) {
+      free(ioctl_buffer);
       return 1;
+    }
     
       errno=EIO;
-      return -1;
+      goto failed;
   }
   
   // copy sector count register (one byte!) to return data
   if (command==CHECK_POWER_MODE)
-    *data=*(char *)&(cmd->tu_command.ata.sector_count);
+    *data=*(char *)&(cmd->ata.sector_count);
   
   // look for nonexistent devices/ports
   if (command==IDENTIFY && !nonempty((unsigned char *)data, 512)) {
     errno=ENODEV;
-    return -1;
+    goto failed;
   }
   
+  free(ioctl_buffer);
   return 0;
+
+failed:
+  free(ioctl_buffer);
+  return -1;
 }
 
-static int get_twe_channel_unit (const char* name, int* unit, int* dev) {
-  // at some  point, we need to figure out which TWE controller any
-  // given disk belongs to.....
-  // at this point, I have no clue how to do this...so for now, it is
-  // always going to be controller 0
-  *dev=0;
-  *unit=0; // not really needed for TWE drives, as we handle that seperately
-  return 0;
+static int get_tw_channel_unit (const char* name, int* unit, int* dev) {
+  char *p;
+
+  /* device node sanity check */
+  for (p = name + 3; *p; p++)
+    if (*p < '0' || *p > '9')
+      return -1;
+  if (strlen(name) > 4 && *(name + 3) == '0')
+    return -1;
+
+  *dev=atoi(name + 3);
+
+  /* no need for unit number */
+  *unit=0;
 }
 
+#ifndef IOCATAREQUEST
 static int get_ata_channel_unit ( const char* name, int* unit, int* dev) {
 #ifndef ATAREQUEST
   *dev=0;
@@ -756,7 +856,7 @@
     return 0;
 #endif
 }
-
+#endif
 
 // Guess device type (ata or scsi) based on device name (FreeBSD
 // specific) SCSI device name in FreeBSD can be sd, sr, scd, st, nst,
@@ -767,7 +867,8 @@
 static const char * fbsd_dev_scsi_tape1 = "sa";
 static const char * fbsd_dev_scsi_tape2 = "nsa";
 static const char * fbsd_dev_scsi_tape3 = "esa";
-static const char * fbsd_dev_twe_disk = "twed";
+static const char * fbsd_dev_twe_ctrl = "twe";
+static const char * fbsd_dev_twa_ctrl = "twa";
 
 static int parse_ata_chan_dev(const char * dev_name, struct freebsd_dev_channel *chan) {
   int len;
@@ -788,11 +889,13 @@
   // form /dev/ad* or ad*
   if (!strncmp(fbsd_dev_ata_disk_prefix, dev_name,
                strlen(fbsd_dev_ata_disk_prefix))) {
+#ifndef IOCATAREQUEST
     if (chan != NULL) {
       if (get_ata_channel_unit(dev_name,&(chan->channel),&(chan->device))<0) {
         return CONTROLLER_UNKNOWN;
       }
     }
+#endif
     return CONTROLLER_ATA;
   }
   
@@ -816,10 +919,20 @@
               strlen(fbsd_dev_scsi_tape3)))
     goto handlescsi;
   
-  if (!strncmp(fbsd_dev_twe_disk,dev_name,
-	       strlen(fbsd_dev_twe_disk))) {
+  if (!strncmp(fbsd_dev_twa_ctrl,dev_name,
+	       strlen(fbsd_dev_twa_ctrl))) {
+    if (chan != NULL) {
+      if (get_tw_channel_unit(dev_name,&(chan->channel),&(chan->device))<0) {
+	return CONTROLLER_UNKNOWN;
+      }
+    }
+    return CONTROLLER_3WARE_9000_CHAR;
+  }
+
+  if (!strncmp(fbsd_dev_twe_ctrl,dev_name,
+	       strlen(fbsd_dev_twe_ctrl))) {
     if (chan != NULL) {
-      if (get_twe_channel_unit(dev_name,&(chan->channel),&(chan->device))<0) {
+      if (get_tw_channel_unit(dev_name,&(chan->channel),&(chan->device))<0) {
 	return CONTROLLER_UNKNOWN;
       }
     }
