--- os_freebsd.h.orig	Sun Sep  5 03:16:07 2004
+++ os_freebsd.h	Fri Aug 19 02:56:17 2005
@@ -21,6 +21,64 @@
  *
  */
 
+/*-
+ * Copyright (c) 2000 Michael Smith
+ * Copyright (c) 2003 Paul Saab
+ * Copyright (c) 2003 Vinod Kashyap
+ * Copyright (c) 2000 BSDi
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ *
+ */
+
+/*
+ * Copyright (c) 2004-05 Applied Micro Circuits Corporation.
+ * Copyright (c) 2004-05 Vinod Kashyap
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ *
+ */
+
 #ifndef OS_FREEBSD_H_
 #define OS_FREEBSD_H_
 
@@ -235,7 +293,6 @@
     u_int32_t	value;
 } TWE_Response_Queue;
 
-
 #endif
 
 #ifdef  HAVE_SYS_TWEIO_H
@@ -269,5 +326,186 @@
    CONTROLLER_3WARE_9000_CHAR -- 9000 controllers via /dev/twa?
 */
 
+#pragma pack(1)
+
+/*
+ * Following cut out of tw_osl_types.h
+ *
+ */
+typedef void			TW_VOID;
+typedef char			TW_INT8;
+typedef unsigned char		TW_UINT8;
+typedef short			TW_INT16;
+typedef unsigned short		TW_UINT16;
+typedef int			TW_INT32;
+typedef unsigned int		TW_UINT32;
+typedef long long		TW_INT64;
+typedef unsigned long long	TW_UINT64;
+
+/*
+ * Following cut out of tw_cl_ioctl.h
+ *
+ */
+/* Driver understandable part of the ioctl packet built by the API. */
+struct tw_cl_driver_packet {
+	TW_UINT32	control_code;
+	TW_UINT32	status;
+	TW_UINT32	unique_id;
+	TW_UINT32	sequence_id;
+	TW_UINT32	os_status;
+	TW_UINT32	buffer_length;
+};
+
+/*
+ * Following cut out of tw_cl_fwif.h
+ *
+ */
+#define TWA_SENSE_DATA_LENGTH		18
+
+/* 7000 structures. */
+struct tw_cl_command_init_connect {
+	TW_UINT8	res1__opcode;	/* 3:5 */
+	TW_UINT8	size;
+	TW_UINT8	request_id;
+	TW_UINT8	res2;
+	TW_UINT8	status;
+	TW_UINT8	flags;
+	TW_UINT16	message_credits;
+	TW_UINT32	features;
+	TW_UINT16	fw_srl;
+	TW_UINT16	fw_arch_id;
+	TW_UINT16	fw_branch;
+	TW_UINT16	fw_build;
+	TW_UINT32	result;
+};
+
+
+/* Structure for downloading firmware onto the controller. */
+struct tw_cl_command_download_firmware {
+	TW_UINT8	sgl_off__opcode;/* 3:5 */
+	TW_UINT8	size;
+	TW_UINT8	request_id;
+	TW_UINT8	unit;
+	TW_UINT8	status;
+	TW_UINT8	flags;
+	TW_UINT16	param;
+	TW_UINT8	sgl[1];
+};
+
+
+/* Structure for hard resetting the controller. */
+struct tw_cl_command_reset_firmware {
+	TW_UINT8	res1__opcode;	/* 3:5 */
+	TW_UINT8	size;
+	TW_UINT8	request_id;
+	TW_UINT8	unit;
+	TW_UINT8	status;
+	TW_UINT8	flags;
+	TW_UINT8	res2;
+	TW_UINT8	param;
+};
+
+
+/* Structure for sending get/set param commands. */
+struct tw_cl_command_param {
+	TW_UINT8	sgl_off__opcode;/* 3:5 */
+	TW_UINT8	size;
+	TW_UINT8	request_id;
+	TW_UINT8	host_id__unit;	/* 4:4 */
+	TW_UINT8	status;
+	TW_UINT8	flags;
+	TW_UINT16	param_count;
+	TW_UINT8	sgl[1];
+};
+
+
+/* Generic command packet. */
+struct tw_cl_command_generic {
+	TW_UINT8	sgl_off__opcode;/* 3:5 */
+	TW_UINT8	size;
+	TW_UINT8	request_id;
+	TW_UINT8	host_id__unit;	/* 4:4 */
+	TW_UINT8	status;
+	TW_UINT8	flags;
+	TW_UINT16	count;	/* block cnt, parameter cnt, message credits */
+};
+
+
+/* Command packet header. */
+struct tw_cl_command_header {
+	TW_UINT8	sense_data[TWA_SENSE_DATA_LENGTH];
+	struct {
+		TW_INT8		reserved[4];
+		TW_UINT16	error;
+		TW_UINT8	padding;
+		TW_UINT8	res__severity;	/* 5:3 */
+	} status_block;
+	TW_UINT8	err_specific_desc[98];
+	struct {
+		TW_UINT8	size_header;
+		TW_UINT16	reserved;
+		TW_UINT8	size_sense;
+	} header_desc;
+};
+
+
+/* 7000 Command packet. */
+union tw_cl_command_7k {
+	struct tw_cl_command_init_connect	init_connect;
+	struct tw_cl_command_download_firmware	download_fw;
+	struct tw_cl_command_reset_firmware	reset_fw;
+	struct tw_cl_command_param		param;
+	struct tw_cl_command_generic		generic;
+	TW_UINT8	padding[1024 - sizeof(struct tw_cl_command_header)];
+};
+
+
+/* 9000 Command Packet. */
+struct tw_cl_command_9k {
+	TW_UINT8	res__opcode;	/* 3:5 */
+	TW_UINT8	unit;
+	TW_UINT16	lun_l4__req_id;	/* 4:12 */
+	TW_UINT8	status;
+	TW_UINT8	sgl_offset; /* offset (in bytes) to sg_list, from the
+					end of sgl_entries */
+	TW_UINT16	lun_h4__sgl_entries;
+	TW_UINT8	cdb[16];
+	TW_UINT8	sg_list[872];/* total struct size =
+					1024-sizeof(cmd_hdr) */
+};
+
+
+/* Full command packet. */
+struct tw_cl_command_packet {
+	struct tw_cl_command_header	cmd_hdr;
+	union {
+		union tw_cl_command_7k	cmd_pkt_7k;
+		struct tw_cl_command_9k cmd_pkt_9k;
+	} command;
+};
+
+/*
+ * Following cut out of tw_osl_ioctl.h
+ *
+ */
+/*
+ * We need the structure below to ensure that the first byte of
+ * data_buf is not overwritten by the kernel, after we return
+ * from the ioctl call.  Note that cmd_pkt has been reduced
+ * to an array of 1024 bytes even though it's actually 2048 bytes
+ * in size.  This is because, we don't expect requests from user
+ * land requiring 2048 (273 sg elements) byte cmd pkts.
+ */
+typedef struct tw_osli_ioctl_no_data_buf {
+	struct tw_cl_driver_packet	driver_pkt;
+	TW_VOID				*pdata; /* points to data_buf */
+	TW_INT8				padding[488 - sizeof(TW_VOID *)];
+	struct tw_cl_command_packet	cmd_pkt;
+} TW_OSLI_IOCTL_NO_DATA_BUF;
+
+#define TW_OSL_IOCTL_FIRMWARE_PASS_THROUGH		\
+	_IOWR('T', 202, TW_OSLI_IOCTL_NO_DATA_BUF)
+
+#pragma pack()
 
 #endif /* OS_FREEBSD_H_ */
